/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 16/05/2007
   derniere modification : 27/06/2007 */


#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEVector.h>
#include <SCE/utils/SCEMatrix.h>
#include <SCE/utils/SCEInert.h>

#include <SCE/core/SCECMatrix.h>
#include <SCE/core/SCECBuffers.h>
#include <SCE/core/SCECTexture.h>
#include <SCE/core/SCECFramebuffer.h>
#include <SCE/core/SCECShader.h>
#include <SCE/core/SCECMaterial.h>

#include <SCE/interface/SCEResources.h>
#include <SCE/interface/SCEMedia.h>
#include <SCE/interface/SCEMesh.h>
#include <SCE/interface/SCETexture.h>
#include <SCE/interface/SCEShaders.h>
#include <SCE/interface/SCEMaterial.h>
#include <SCE/interface/SCESMLoader.h>

#include <SCE/interface/SCEModel.h>



/* type SM (Sce Mesh) */
static int sce_sm_media_type;


int SCE_Model_Init(void)
{
    static int is_init = SCE_FALSE;
    
    if(!is_init)
    {
        /* enregistrement des loaders par defaut */
        /* SM */
        sce_sm_media_type = SCE_Media_GenTypeID();
        SCE_Media_RegisterLoader(sce_sm_media_type, SCE_SM_MAGIC,
                                 ".sm", SCE_SM_Load);
        
        /* ... (obj a venir) */
        
        is_init = SCE_TRUE;
    }
    
    return (SCE_OK);
}

int SCE_Model_Quit(void)
{
    return (SCE_OK);
}


/* revise le 27/06/2007 */
SCE_SModel* SCE_Model_Create(const int animated)
{
    unsigned int i;
    SCE_SModel *mdl = NULL;
    
    mdl = SCE_malloc(sizeof *mdl);
    if(mdl == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    mdl->mesh_lod = 0;
    
    mdl->mat = NULL;
    for(i=0; i<SCE_MAX_GEOMETRY_LOD; i++)
        mdl->mesh[i] = NULL;
    
    SCE_Vector3_Init(mdl->pos, 0.0f);
    SCE_Vector3_Init(mdl->axe, 0.0f);
    SCE_Vector3_Init(mdl->front, 0.0f); mdl->front[0] = 1.0f;
    
    SCE_Inert_Init(&mdl->angle);
    SCE_Inert_Init(&mdl->front_mv);
    SCE_Inert_Init(&mdl->scale_x);
    SCE_Inert_Init(&mdl->scale_y);
    SCE_Inert_Init(&mdl->scale_z);
    SCE_Inert_Init(&mdl->angle_x);
    SCE_Inert_Init(&mdl->angle_y);
    SCE_Inert_Init(&mdl->angle_z);
    
    SCE_Inert_Set(&mdl->scale_x, 1.0f);
    SCE_Inert_Set(&mdl->scale_y, 1.0f);
    SCE_Inert_Set(&mdl->scale_z, 1.0f);
    
    SCE_Inert_Accum(&mdl->front_mv, SCE_FALSE);
    
    mdl->affect_front = SCE_TRUE;
    mdl->animated = animated;
    
    return (mdl);
}

/* revise le 23/06/2007 */
void SCE_Model_Delete(SCE_SModel *mdl)
{
    unsigned int i;
    
    if(mdl != NULL)
    {
        for(i=0; i<SCE_MAX_GEOMETRY_LOD; i++)
            SCE_Mesh_Delete(mdl->mesh[i]);
        SCE_Material_Delete(mdl->mat);
        SCE_free(mdl);
    }
}


/* revise le 27/06/2007 */
void SCE_Model_SetPos(SCE_SModel *mdl, float x, float y, float z)
{
    if(mdl != NULL)
    {
        mdl->pos[0] = x;
        mdl->pos[1] = y;
        mdl->pos[2] = z;
    }
}

/* ajoute le 27/06/2007 */
void SCE_Model_SetPosv(SCE_SModel *mdl, SCE_TVector3 vec)
{
    if(mdl != NULL)
    {
        SCE_Vector3_Operator1v(mdl->pos, =, vec);
    }
}

/* revise le 27/06/2007 */
void SCE_Model_SetAxe(SCE_SModel *mdl, float x, float y, float z)
{
    if(mdl != NULL)
    {
        mdl->axe[0] = x;
        mdl->axe[1] = y;
        mdl->axe[2] = z;
    }
}

/* ajoute le 27/06/2007 */
void SCE_Model_SetAxev(SCE_SModel *mdl, SCE_TVector3 vec)
{
    if(mdl != NULL)
    {
        SCE_Vector3_Operator1v(mdl->axe, =, vec);
    }
}

/* ajoute le 27/06/2007 */
void SCE_Model_Rotate(SCE_SModel *mdl, float a, float ax, float ay, float az)
{
    if(mdl != NULL)
    {
        SCE_Inert_Set(&mdl->angle, a);
        SCE_Inert_Set(&mdl->angle_x, ax);
        SCE_Inert_Set(&mdl->angle_y, ay);
        SCE_Inert_Set(&mdl->angle_z, az);
    }
}

/* ajoute le 27/06/2007 */
void SCE_Model_AffectFrontVec(SCE_SModel *mdl, int affect)
{
    if(mdl != NULL)
        mdl->affect_front = affect;
}


/* revise le 27/06/2007 */
void SCE_Model_ComputeMovs(SCE_SModel *mdl)
{
    SCE_TMatrix4 m;
    
    if(mdl != NULL)
    {
        /* calcul des variables d'inertie */
        SCE_Inert_ComputeVar(&mdl->angle);
        SCE_Inert_ComputeVar(&mdl->front_mv);
        SCE_Inert_ComputeVar(&mdl->scale_x);
        SCE_Inert_ComputeVar(&mdl->scale_y);
        SCE_Inert_ComputeVar(&mdl->scale_z);
        SCE_Inert_ComputeVar(&mdl->angle_x);
        SCE_Inert_ComputeVar(&mdl->angle_y);
        SCE_Inert_ComputeVar(&mdl->angle_z);
        
        
        if(mdl->affect_front)
        {
            /* application des rotations d'angles independants
            au vecteur "avant" du modele */
            /* x */
            SCE_Matrix4_LoadIdentity(m);
            SCE_Matrix4_RotX(m, SCE_Inert_Get(&mdl->angle_x));
            SCE_Matrix4_MulV3(m, mdl->front);
            /* y */
            SCE_Matrix4_LoadIdentity(m);
            SCE_Matrix4_RotY(m, SCE_Inert_Get(&mdl->angle_y));
            SCE_Matrix4_MulV3(m, mdl->front);
            /* z */
            SCE_Matrix4_LoadIdentity(m);
            SCE_Matrix4_RotZ(m, SCE_Inert_Get(&mdl->angle_z));
            SCE_Matrix4_MulV3(m, mdl->front);
            
            /* application de la rotation autour du vecteur "axe" */
            SCE_Matrix4_LoadIdentity(m);
            SCE_Matrix4_Rotatev(m, SCE_Inert_Get(&mdl->angle), mdl->axe);
            SCE_Matrix4_MulV3(m, mdl->front);
            SCE_Vector3_Normalize(mdl->front);
        }
        
        /* mouvement frontal */
        SCE_Vector3_Operator2(mdl->pos, +=,
                              mdl->front, *, SCE_Inert_Get(&mdl->front_mv));
        
        /* ... NOTE */
    }
}


/* revise le 27/06/2007 */
void SCE_Model_PushMatrices(SCE_SModel *mdl)
{
    if(mdl != NULL)
    {
        /* push */
        SCE_CPushSetMatrix(SCE_MAT_TEXTURE);
        SCE_CPushSetMatrix(SCE_MAT_MODELVIEW);
        
        SCE_CTranslateMatrixv(mdl->pos);
        SCE_CRotateMatrixv(SCE_Inert_Get(&mdl->angle), mdl->axe);
        SCE_CRotateMatrix(SCE_Inert_Get(&mdl->angle_x), 1.0f, 0.0f, 0.0f);
        SCE_CRotateMatrix(SCE_Inert_Get(&mdl->angle_y), 0.0f, 1.0f, 0.0f);
        SCE_CRotateMatrix(SCE_Inert_Get(&mdl->angle_z), 0.0f, 0.0f, 1.0f);
        SCE_CScaleMatrix(SCE_Inert_Get(&mdl->scale_x),
                         SCE_Inert_Get(&mdl->scale_y),
                         SCE_Inert_Get(&mdl->scale_z));
    }
}

/* revise le 14/06/2007 */
void SCE_Model_PopMatrices(void)
{
    SCE_CPopSetMatrix(SCE_MAT_TEXTURE);
    SCE_CPopSetMatrix(SCE_MAT_MODELVIEW);
}


/* ajoute le 14/06/2007 */
void SCE_Model_Bind(SCE_SModel *mdl)
{
    if(mdl != NULL)
    {
        SCE_Mesh_Bind(mdl->mesh[mdl->mesh_lod]);
        SCE_Material_Use(mdl->mat);
    }
    else
    {
        SCE_Material_Use(NULL);
        SCE_Mesh_Bind(NULL);
    }
}

/* ajoute le 14/06/2007 */
void SCE_Model_BindMesh(SCE_SModel *mdl)
{
    if(mdl != NULL)
        SCE_Mesh_Bind(mdl->mesh[mdl->mesh_lod]);
    else
        SCE_Mesh_Bind(NULL);
}

/* ajoute le 14/06/2007 */
void SCE_Model_BindMaterial(SCE_SModel *mdl)
{
    if(mdl != NULL)
        SCE_Material_Use(mdl->mat);
    else
        SCE_Material_Use(NULL);
}

/* ajoute le 14/06/2007 */
void SCE_Model_Draw(void)
{
    SCE_Mesh_Draw();
}


void SCE_Model_Render(SCE_SModel *mdl)
{
    if(mdl != NULL)
    {
        SCE_Model_ComputeMovs(mdl);
        
        /* modifications matricielles */
        SCE_Model_PushMatrices(mdl);
        
        SCE_Material_Use(mdl->mat);
        SCE_Mesh_Render(0, 1, &mdl->mesh[mdl->mesh_lod]);
        SCE_Material_Use(NULL);
        
        /* restauration des matrices */
        SCE_Model_PopMatrices();
    }
}


/* ajoute le 25/06/2007 */
/* revise le 27/06/2007 */
int SCE_Model_LoadMesh(SCE_SModel *mdl, const char *fname, int use_lod)
{
    SCE_SMesh **meshs = NULL;
    int nb_meshs, i;/*
    unsigned int first_free;*/ /* premier mesh libre dans le modele */
    
    
    for(i=0; i<SCE_MAX_GEOMETRY_LOD; i++)
    
    /* chargement du media */
    meshs = SCE_Resource_Load(fname, &nb_meshs, NULL);
    if(meshs == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    
    /* on verifie s'il n'y a pas trop de LODs dans le mesh */
    if(nb_meshs > SCE_MAX_GEOMETRY_LOD && use_lod)
    {
        Logger_Log(SCE_BAD_FORMAT);
        Logger_LogMsg("LOD (Level Of Detail) of '%s' is %d, but the maximum is "
                      "%d", fname, nb_meshs, SCE_MAX_GEOMETRY_LOD);
        
        for(i=nb_meshs; i!=0; i--)
            SCE_Mesh_Delete(meshs[i]);
        SCE_free(meshs);
        
        return (SCE_ERROR);
    }
    
    /** TODO il faut verifier si le modele ne possede pas deja des meshs... */
    
    
    
    /* transfert dans le modele */
    for(i=0; i<nb_meshs; i++)
        mdl->mesh[i] = meshs[i];
    
    /* meshs doit IMPERATIVEMENT etre un tableau de pointeurs sur SCE_SMesh */
    SCE_free(meshs);
    
    return (SCE_OK);
}


/* ajoute le 25/06/2007 */
void SCE_Model_DeleteLOD(SCE_SModel *mdl, SCEuint lod)
{
    if(mdl != NULL)
    {
        if(lod < SCE_MAX_GEOMETRY_LOD)
            SCE_Mesh_Delete(mdl->mesh[lod]);
    }
}
